/*global module require */
"use strict";

const util = require("./utilities");
const request = require("./requestPromise");

/**
 * @async
 * @description This function will get an already created report.
 * @param {string} [accessToken="null access token"]
 * @param {string} reportUUIDlistVoicemails
 * @param {string} templateUUID
 * @param {object} [trace = {}] - optional microservice lifecycle trace headers 
 * @returns {Promise<object>} - Promise resolving to a data object containing new relationship
 */
const getReport = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null access token";
  let reportUUID = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "null report uuid";
  let templateUUID = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : "null template uuid";
  let trace = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};
  try {
    const MS = util.getEndpoint("activity");
    const requestOptions = {
      method: "GET",
      uri: `${MS}/report`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      qs: {
        report_uuid: reportUUID,
        template_uuid: templateUUID
      },
      resolveWithFullResponse: true,
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    if (response.statusCode === 204) {
      return {};
    }
    return response.body;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function will list available report templates.
 * @param {string} [accessToken="null access token"]
 * @param {number} [offset = 0] - pagination offset
 * @param {number} [limit = 10] = pagination limit
 * @param {object} [trace = {}] - optional microservice lifecycle trace headers 
 * @returns {Promise<object>} - Promise resolving to a data object containing new relationship
 */
const listReportTemplates = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null access token";
  let offset = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0;
  let limit = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 10;
  let trace = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};
  try {
    const MS = util.getEndpoint("activity");
    const requestOptions = {
      method: "GET",
      uri: `${MS}/report/template`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      qs: {
        offset: offset,
        limit: limit
      },
      resolveWithFullResponse: false,
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function will create a report based on a template.
 * @param {string} [accessToken="null access token"] - access token for cpaas systems
 * @param {string} [templateUuid="null account uuid"] - report template uuid
 * @param {string} [ownerUuid=undefined] - owner uuid (CPaaS user_uuid) - required if account uuid not specified
 * @param {string} [accountUuid=undefined] - CPaaS account uuid - required if owner uuid not specified
 * @param {object} [paramaters={}] - report parameters
 * @param {object} [trace={}] - optional CPaaS microservice lifecycle headers
 * @returns {Promise<object>} - Promise resolving to an object containing report uuid
 */
const runReport = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null access token";
  let templateUuid = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "null account uuid";
  let ownerUuid = arguments.length > 2 ? arguments[2] : undefined;
  let accountUuid = arguments.length > 3 ? arguments[3] : undefined;
  let parameters = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : {};
  let trace = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : {};
  try {
    const MS = util.getEndpoint("activity");
    const requestOptions = {
      method: "POST",
      uri: `${MS}/report/${templateUuid}`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      qs: {},
      body: {
        parameters: parameters
      },
      json: true
    };
    if (typeof ownerUuid !== "undefined") {
      requestOptions.qs.owner_uuid = ownerUuid;
    } else if (typeof accountUuid !== "undefined") {
      requestOptions.qs.account_uuid = accountUuid;
    } else {
      throw {
        "code": 400,
        "message": "request requires either owner or account uuid"
      };
    }
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function will list registered activity types
 * @param {string} [accessToken="null access token"]
 * @param {string} [include=undefined] - optional include 
 * @param {number} [offset = 0] - pagination offset
 * @param {number} [limit = 10] = pagination limit
 * @param {object} [trace = {}] - optional microservice lifecycle trace headers 
 * @returns {Promise<object>} - Promise resolving to a data object containing new relationship
 */
const listRegisteredTypes = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null access token";
  let include = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : undefined;
  let offset = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 0;
  let limit = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : 10;
  let trace = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : {};
  try {
    const MS = util.getEndpoint("activity");
    const requestOptions = {
      method: "GET",
      uri: `${MS}/register`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      qs: {
        offset,
        limit,
        include
      },
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function will list registered sub activity types
 * @param {string} [accessToken="null access token"]
 * @param {string} [include=undefined] - optional include 
 * @param {number} [offset = 0] - pagination offset
 * @param {number} [limit = 10] = pagination limit
 * @param {object} [trace = {}] - optional microservice lifecycle trace headers 
 * @returns {Promise<object>} - Promise resolving to a data object containing new relationship
 */
const listRegisteredSubTypes = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null access token";
  let include = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : undefined;
  let offset = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 0;
  let limit = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : 10;
  let trace = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : {};
  try {
    const MS = util.getEndpoint("activity");
    const requestOptions = {
      method: "GET",
      uri: `${MS}/register/subtype`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      qs: {
        offset,
        limit,
        include
      },
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function will register an activity type
 * @param {string} [accessToken="null access token"]
 * @param {string} [activity_type = ""] - activity type
 * @param {number} [version = 1] = version
 * @param {object} [schema={}] = schema
 * @param {object} [trace = {}] - optional microservice lifecycle trace headers 
 * @returns {Promise<object>} - Promise resolving to a data object containing new relationship
 */
const registerType = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null access token";
  let activity_type = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : undefined;
  let version = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 1;
  let schema = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : undefined;
  let trace = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : {};
  try {
    const MS = util.getEndpoint("activity");
    const requestOptions = {
      method: "POST",
      uri: `${MS}/register`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      body: {
        activity_type,
        version,
        schema
      },
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function will register a sub activity type
 * @param {string} [accessToken="null access token"]
 * @param {string} [activity_type = ""] - activity type
 * @param {string} [activity_subtype = ""] - activity subtype
 * @param {number} [version = 1] = version
 * @param {object} [schema={}] = schema
 * @param {object} [trace = {}] - optional microservice lifecycle trace headers 
 * @returns {Promise<object>} - Promise resolving to a data object containing new relationship
 */
const registerSubType = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null access token";
  let activity_type = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : undefined;
  let activity_subtype = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : undefined;
  let version = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : 1;
  let schema = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : undefined;
  let trace = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : {};
  try {
    const MS = util.getEndpoint("activity");
    const requestOptions = {
      method: "POST",
      uri: `${MS}/register/subtype`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      body: {
        activity_type,
        activity_subtype,
        version,
        schema
      },
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function will create an activity report template
 * @param {string} [accessToken="null access token"]
 * @param {string} [name = ""] - name
 * @param {string} [description = ""] - description
 * @param {string} [default_view = ""] = default view
 * @param {object} [data_source={}] = data_source
 * @param {array} [parameters=[]] = parameters
 * @param {array} [columns=[]] = columns
 * @param {array} [limits=[]] = limits
 * @param {object} [sort={}] = sort
 * @param {array} [aggregates=[]] = aggregates
 * @param {object} [visualizations={}] = visualizations
 * @param {object} [trace = {}] - optional microservice lifecycle trace headers 
 * @returns {Promise<object>} - Promise resolving to an object 
 */
const createReportTemplate = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null access token";
  let name = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : undefined;
  let description = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : undefined;
  let default_view = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : undefined;
  let data_source = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : undefined;
  let parameters = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : undefined;
  let columns = arguments.length > 6 && arguments[6] !== undefined ? arguments[6] : undefined;
  let limits = arguments.length > 7 && arguments[7] !== undefined ? arguments[7] : undefined;
  let sort = arguments.length > 8 && arguments[8] !== undefined ? arguments[8] : undefined;
  let aggregates = arguments.length > 9 && arguments[9] !== undefined ? arguments[9] : undefined;
  let visualizations = arguments.length > 10 && arguments[10] !== undefined ? arguments[10] : undefined;
  let trace = arguments.length > 11 && arguments[11] !== undefined ? arguments[11] : {};
  try {
    const MS = util.getEndpoint("activity");
    const body = {
      name,
      description,
      default_view,
      data_source,
      parameters,
      columns,
      limits,
      sort,
      aggregates,
      visualizations
    };
    const requestOptions = {
      method: "POST",
      uri: `${MS}/report/template`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      body: body,
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function will update an activity report template
 * @param {string} [accessToken="null access token"]
 * @param {string} [template_uuid = ""] - template_uuid
 * @param {object} [template={}] = template
 * @param {object} [trace = {}] - optional microservice lifecycle trace headers 
 * @returns {Promise<object>} - Promise resolving to an object 
 */
const updateReportTemplate = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null access token";
  let template_uuid = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : undefined;
  let template = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : undefined;
  let trace = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};
  try {
    const MS = util.getEndpoint("activity");
    const requestOptions = {
      method: "PUT",
      uri: `${MS}/report/template/${template_uuid}`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      body: template,
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function will delete an activity report template
 * @param {string} [accessToken="null access token"]
 * @param {string} [template_uuid = ""] - template_uuid
 * @param {object} [trace = {}] - optional microservice lifecycle trace headers 
 * @returns {Promise<object>} - Promise resolving to an object 
 */
const deleteReportTemplate = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null access token";
  let template_uuid = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : undefined;
  let trace = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  try {
    const MS = util.getEndpoint("activity");
    const requestOptions = {
      method: "DELETE",
      uri: `${MS}/report/template/${template_uuid}`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function will list activites based on parameters .
 * @param {string} [accessToken="null access token"]
 * @param {object} [filters={}}] - optional object consisting of key and value to use in url as parameters.
 * @param {number} [offset = 0] - pagination offset
 * @param {number} [limit = 10] = pagination limit
 * @param {object} [trace = {}] - optional microservice lifecycle trace headers 
 * @returns {Promise<object>} - Promise resolving to a data object of activities
 */
const listActivities = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null access token";
  let filters = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
  let offset = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 0;
  let limit = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : 10;
  let trace = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : {};
  try {
    const MS = util.getEndpoint("activity");
    const requestOptions = {
      method: "GET",
      uri: `${MS}/activity`,
      qs: {
        offset: offset,
        limit: limit
      },
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    if (filters) {
      Object.keys(filters).forEach(keyName => {
        requestOptions.qs[keyName] = filters[keyName];
      });
    }
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};
module.exports = {
  createReportTemplate,
  deleteReportTemplate,
  getReport,
  listActivities,
  listRegisteredTypes,
  listRegisteredSubTypes,
  listReportTemplates,
  registerType,
  registerSubType,
  runReport,
  updateReportTemplate
};