/* global require module*/
"use strict";

const util = require("./utilities");
const request = require("./requestPromise");
const logger = require("./node-logger").getInstance();

/**
 * @async
 * @description - This function will assign numbers a SMS 10DLC Campaign
 * @param {string} [accessToken="null accessToken"] cpaas access token
 * @param {string} [accountUUID="null accountUUID"] cpaas account uuid to provision
 * @param {string} [campaignId="null campaignId"] 10dlc campaign id
 * @param {array} [numbers=[]] array of numbers to assign
 * @param {object} [trace={}] - microservice lifecyce headers
 * @returns {Promise} - Promise resolving to a object confirming the assignment of numbers
 */
const assignToSMSCampaign = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let accountUUID = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "null accountUUID";
  let campaignId = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : [];
  let numbers = arguments.length > 3 ? arguments[3] : undefined;
  let trace = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : {};
  try {
    const MS = util.getEndpoint("campaigns");
    const requestOptions = {
      method: "POST",
      uri: `${MS}/10dlc/customers/${accountUUID}/campaigns/${campaignId}/numbers/assign`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "x-api-version": `${util.getVersion()}`,
        "Content-type": "application/json"
      },
      body: {
        tns: numbers
      },
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description - This function will create a SMS 10DLC Brand
 * @param {string} [accessToken="null accessToken"] - cpaas access token
 * @param {string} [accountUUID] - account uuid to provision on behalf of
 * @param {string} legalName company legal name
 * @param {string} brandName brand name
 * @param {string} organizationType organization type
 * @param {string} [registrationCountry="US"] registration country
 * @param {string} taxId tax id
 * @param {string} [taxIdCountry="US"] tax country
 * @param {string} altBusinessIdType alternate business type
 * @param {string} altBusinessId alternate business id
 * @param {string} vertical business vertical
 * @param {string} address legal streed address
 * @param {string} city legal address city
 * @param {string} state legal address state
 * @param {string} postalCode legal address postal code
 * @param {string} website company website
 * @param {string} stockSymbol public stock symbol
 * @param {string} stockExchange public stock exchange
 * @param {string} emailAddress contact email address
 * @param {string} business_contact_email business contact email address
 * @param {string} phoneNumber contact phone number
 * @param {string} firstName contact first name
 * @param {string} lastName contact last name
 * @param {object} [trace={}] - microservice lifecyce headers
 * @returns {Promise} - Promise resolving to a object containing a new 10DLC Brand
 */
const createBrand = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let accountUUID = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "null accountUUID";
  let legalName = arguments.length > 2 ? arguments[2] : undefined;
  let brandName = arguments.length > 3 ? arguments[3] : undefined;
  let organizationType = arguments.length > 4 ? arguments[4] : undefined;
  let registrationCountry = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : "US";
  let taxId = arguments.length > 6 ? arguments[6] : undefined;
  let taxIdCountry = arguments.length > 7 && arguments[7] !== undefined ? arguments[7] : "US";
  let altBusinessIdType = arguments.length > 8 ? arguments[8] : undefined;
  let altBusinessId = arguments.length > 9 ? arguments[9] : undefined;
  let vertical = arguments.length > 10 ? arguments[10] : undefined;
  let address = arguments.length > 11 ? arguments[11] : undefined;
  let city = arguments.length > 12 ? arguments[12] : undefined;
  let state = arguments.length > 13 ? arguments[13] : undefined;
  let postalCode = arguments.length > 14 ? arguments[14] : undefined;
  let website = arguments.length > 15 ? arguments[15] : undefined;
  let stockSymbol = arguments.length > 16 ? arguments[16] : undefined;
  let stockExchange = arguments.length > 17 ? arguments[17] : undefined;
  let emailAddress = arguments.length > 18 ? arguments[18] : undefined;
  let businessContactEmail = arguments.length > 19 ? arguments[19] : undefined;
  let phoneNumber = arguments.length > 20 ? arguments[20] : undefined;
  let firstName = arguments.length > 21 ? arguments[21] : undefined;
  let lastName = arguments.length > 22 ? arguments[22] : undefined;
  let trace = arguments.length > 23 && arguments[23] !== undefined ? arguments[23] : {};
  try {
    const MS = util.getEndpoint("campaigns");
    const requestOptions = {
      method: "POST",
      uri: `${MS}/10dlc/customers`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "x-api-version": `${util.getVersion()}`,
        "Content-type": "application/json"
      },
      body: {
        legal_name: legalName,
        brand_name: brandName,
        organization_type: organizationType,
        registration_country: registrationCountry,
        e_i_n: taxId,
        e_i_n_country: taxIdCountry,
        alt_business_id_type: altBusinessIdType,
        alt_business_id: altBusinessId,
        vertical: vertical,
        address: address,
        city: city,
        state: state,
        postal_code: postalCode,
        website: website,
        stock_symbol: stockSymbol,
        stock_exchange: stockExchange,
        email_address: emailAddress,
        business_contact_email: businessContactEmail,
        phone_number: phoneNumber,
        first_name: firstName,
        last_name: lastName,
        account_uuid: accountUUID
      },
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function will register a campaign for a CPaaS account that has an active brand registration
 * @param {string} [accessToken="null accessToken"] CPaaS access token
 * @param {string} [accountUUID="null accountUUID"] CPaaS account uuid
 * @param {string} description summary description of this campaign between 40 and 4096 characters
 * @param {string} displayName a friendly name for the campaign
 * @param {string} sampleMsg1 sample message 1
 * @param {string} sampleMsg2 sample message 2
 * @param {string} sampleMsg3 sample message 3
 * @param {string} sampleMsg4 sample message 4
 * @param {string} sampleMsg5 sample message 5
 * @param {boolean} numPooling campaign utilize pool of phone numbers?
 * @param {boolean} directLending messages used for lending or loan arangements?
 * @param {boolean} embeddedLink messages contain embedded URLs?
 * @param {boolean} embeddedPhone messages contain embedded phone numbers?
 * @param {boolean} affMarketing message content controlled by affiliate marketing other than the brand?
 * @param {boolean} ageGated campaign includes age gated message content?
 * @param {string} usecaseId usecase id
 * @param {array} subUsecaseIds sub usecase ids (strings)
 * @param {string} [messageFlow="Users may opt-in by sending START to any number associated with the campaign. Users may also sign up to receive messages from this campaign via a website after accepting terms and conditions."] how subscribers are added or opt-in to the campaign
 * @param {string} termsLink url for viewing terms and conditions
 * @param {string} privacyLink url for viewing privacy policy
 * @param {boolean} [autoRenewal=true] subscription auto-renews?
 * @param {boolean} [subOptIn=true] provides automated opt-in?
 * @param {string} [optInKeywords="START"] opt-in keywords (comma separated)
 * @param {string} [optInMessage="You have replied \"START\" and will begin receiving messages again. Reply \"STOP\" to unsubscribe."] automated opt-in message
 * @param {boolean} [subOptOut=true] provides automated opt-out?
 * @param {string} [optOutKeywords="STOP"] opt-out key words (comma separated)
 * @param {string} [optOutMessage="You have been unsubscribed from the mailing list and will not receive any more messages. Send \"START\" to resubscribe."] automated opt-out message
 * @param {boolean} [subHelp=true] provides automated help?
 * @param {string} [helpKeywords="HELP"] help keywords (comma separated)
 * @param {string} [helpMessage="To unsubscribe, reply \"STOP\" to this number.  Msg&data rates may apply."] automated help message
 * @param {object} [trace={}] optional CPaaS lifecycle headers
 * @return {Promsise<object>} promise resolving to a brand registration document
 */
const createCampaign = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let accountUUID = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "null accountUUID";
  let description = arguments.length > 2 ? arguments[2] : undefined;
  let displayName = arguments.length > 3 ? arguments[3] : undefined;
  let sampleMsg1 = arguments.length > 4 ? arguments[4] : undefined;
  let sampleMsg2 = arguments.length > 5 ? arguments[5] : undefined;
  let sampleMsg3 = arguments.length > 6 ? arguments[6] : undefined;
  let sampleMsg4 = arguments.length > 7 ? arguments[7] : undefined;
  let sampleMsg5 = arguments.length > 8 ? arguments[8] : undefined;
  let numPooling = arguments.length > 9 ? arguments[9] : undefined;
  let directLending = arguments.length > 10 ? arguments[10] : undefined;
  let embeddedLink = arguments.length > 11 ? arguments[11] : undefined;
  let embeddedPhone = arguments.length > 12 ? arguments[12] : undefined;
  let affMarketing = arguments.length > 13 ? arguments[13] : undefined;
  let ageGated = arguments.length > 14 ? arguments[14] : undefined;
  let usecaseId = arguments.length > 15 ? arguments[15] : undefined;
  let subUsecaseIds = arguments.length > 16 && arguments[16] !== undefined ? arguments[16] : [];
  let messageFlow = arguments.length > 17 && arguments[17] !== undefined ? arguments[17] : "Users may opt-in by sending START to any number associated with the campaign. Users may also sign up to receive messages from this campaign via a website after accepting terms and conditions.";
  let termsLink = arguments.length > 18 ? arguments[18] : undefined;
  let privacyLink = arguments.length > 19 ? arguments[19] : undefined;
  let autoRenewal = arguments.length > 20 && arguments[20] !== undefined ? arguments[20] : true;
  let subOptIn = arguments.length > 21 && arguments[21] !== undefined ? arguments[21] : true;
  let optInKeywords = arguments.length > 22 && arguments[22] !== undefined ? arguments[22] : "START";
  let optInMessage = arguments.length > 23 && arguments[23] !== undefined ? arguments[23] : 'You have replied "START" and will begin receiving messages again. Reply "STOP" to unsubscribe.';
  let subOptOut = arguments.length > 24 && arguments[24] !== undefined ? arguments[24] : true;
  let optOutKeywords = arguments.length > 25 && arguments[25] !== undefined ? arguments[25] : "STOP";
  let optOutMessage = arguments.length > 26 && arguments[26] !== undefined ? arguments[26] : 'You have been unsubscribed from the mailing list and will not receive any more messages. Send "START" to resubscribe.';
  let subHelp = arguments.length > 27 && arguments[27] !== undefined ? arguments[27] : true;
  let helpKeywords = arguments.length > 28 && arguments[28] !== undefined ? arguments[28] : "HELP";
  let helpMessage = arguments.length > 29 && arguments[29] !== undefined ? arguments[29] : 'To unsubscribe, reply "STOP" to this number.  Msg&data rates may apply.';
  let trace = arguments.length > 30 && arguments[30] !== undefined ? arguments[30] : {};
  try {
    const MS = util.getEndpoint("campaigns");
    const requestOptions = {
      method: "POST",
      uri: `${MS}/10dlc/customers/${accountUUID}/campaigns`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "x-api-version": `${util.getVersion()}`,
        "Content-type": "application/json"
      },
      body: {
        description: description,
        display_name: displayName,
        message_flow: messageFlow,
        sample_msg1: sampleMsg1,
        sample_msg2: sampleMsg2,
        sample_msg3: sampleMsg3,
        sample_msg4: sampleMsg4,
        sample_msg5: sampleMsg5,
        terms_link: termsLink,
        privacy_link: privacyLink,
        sub_opt_in: subOptIn,
        sub_opt_out: subOptOut,
        sub_help: subHelp,
        num_pooling: numPooling,
        direct_lending: directLending,
        embedded_link: embeddedLink,
        embedded_phone: embeddedPhone,
        aff_marketing: affMarketing,
        age_gated: ageGated,
        usecase_id: usecaseId,
        sub_usecase_ids: subUsecaseIds,
        help_message: helpMessage,
        auto_renewal: autoRenewal,
        opt_in_keywords: optInKeywords,
        opt_out_keywords: optOutKeywords,
        help_keywords: helpKeywords,
        opt_in_message: optInMessage,
        opt_out_message: optOutMessage
      },
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description - This function will delete a SMS 10DLC Brand
 * @param {string} [accessToken="null accessToken"] - cpaas access token
 * @param {string} [accountUUID] - account uuid to provision on behalf of
 * @param {object} [trace={}] - microservice lifecyce headers
 * @returns {Promise} - Promise resolving to a object confirming the delete request was received
 */
const deleteBrand = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let accountUUID = arguments.length > 1 ? arguments[1] : undefined;
  let trace = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  try {
    const MS = util.getEndpoint("campaigns");
    const requestOptions = {
      method: "DELETE",
      uri: `${MS}/10dlc/customers/${accountUUID}`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "x-api-version": `${util.getVersion()}`,
        "Content-type": "application/json"
      },
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description - This function will delete a campaign from a SMS 10DLC Brand / CPaaS account
 * @param {string} [accessToken="null accessToken"] - cpaas access token
 * @param {string} [accountUUID] - CPaaS account uuid
 * @param {string} [campaignId="null campaignId"] 10DLC campaign id to retrieve
 * @param {object} [trace={}] - microservice lifecyce headers
 * @returns {Promise} - Promise resolving to a object containing a CPaaS account 10DLC brand registrations
 */
const deleteCampaign = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let accountUUID = arguments.length > 1 ? arguments[1] : undefined;
  let campaignId = arguments.length > 2 ? arguments[2] : undefined;
  let trace = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};
  try {
    const MS = util.getEndpoint("campaigns");
    const requestOptions = {
      method: "DELETE",
      uri: `${MS}/10dlc/customers/${accountUUID}/campaigns/${campaignId}`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "x-api-version": `${util.getVersion()}`,
        "Content-type": "application/json"
      },
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description - This function will get a SMS 10DLC Brand
 * @param {string} [accessToken="null accessToken"] - cpaas access token
 * @param {string} [accountUUID] - account uuid to provision on behalf of
 * @param {object} [trace={}] - microservice lifecyce headers
 * @returns {Promise} - Promise resolving to a object containing a CPaaS account 10DLC brand registrations
 */
const getBrand = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let accountUUID = arguments.length > 1 ? arguments[1] : undefined;
  let trace = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  try {
    const MS = util.getEndpoint("campaigns");
    const requestOptions = {
      method: "GET",
      uri: `${MS}/10dlc/customers/${accountUUID}`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "x-api-version": `${util.getVersion()}`,
        "Content-type": "application/json"
      },
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description - This function will get campaign options for a SMS 10DLC Brand
 * @param {string} [accessToken="null accessToken"] - cpaas access token
 * @param {string} [accountUUID] - account uuid to provision on behalf of
 * @param {object} [trace={}] - microservice lifecyce headers
 * @returns {Promise} - Promise resolving to a object containing a CPaaS account 10DLC brand registrations
 */
const getBrandUseCases = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let accountUUID = arguments.length > 1 ? arguments[1] : undefined;
  let trace = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  try {
    const MS = util.getEndpoint("campaigns");
    const requestOptions = {
      method: "GET",
      uri: `${MS}/10dlc/customers/${accountUUID}/usecases`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "x-api-version": `${util.getVersion()}`,
        "Content-type": "application/json"
      },
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description - This function will get a campaign for a SMS 10DLC Brand / CPaaS account
 * @param {string} [accessToken="null accessToken"] - cpaas access token
 * @param {string} [accountUUID] - CPaaS account uuid
 * @param {string} [campaignId="null campaignId"] 10DLC campaign id to retrieve
 * @param {object} [trace={}] - microservice lifecyce headers
 * @returns {Promise} - Promise resolving to a object containing a CPaaS account 10DLC brand registrations
 */
const getCampaign = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let accountUUID = arguments.length > 1 ? arguments[1] : undefined;
  let campaignId = arguments.length > 2 ? arguments[2] : undefined;
  let trace = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};
  try {
    const MS = util.getEndpoint("campaigns");
    const requestOptions = {
      method: "GET",
      uri: `${MS}/10dlc/customers/${accountUUID}/campaigns/${campaignId}`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "x-api-version": `${util.getVersion()}`,
        "Content-type": "application/json"
      },
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description - This function will deprovision numbers from a CPaaS account and/or user
 * @param {string} [accessToken="null accessToken"] - cpaas access token
 * @param {object} [trace={}] - microservice lifecyce headers
 * @returns {Promise} - Promise resolving to a object containing deprovisioning confirmation details
 */
const getEnumerations = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let trace = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
  try {
    const MS = util.getEndpoint("campaigns");
    const requestOptions = {
      method: "GET",
      uri: `${MS}/10dlc/help/get_enumerations`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "x-api-version": `${util.getVersion()}`,
        "Content-type": "application/json"
      },
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description - This function will list campaigns for a SMS 10DLC Brand / CPaaS account
 * @param {string} [accessToken="null accessToken"] - cpaas access token
 * @param {string} [accountUUID] - CPaaS account uuid
 * @param {object} [trace={}] - microservice lifecyce headers
 * @returns {Promise} - Promise resolving to a object containing a CPaaS account 10DLC brand registrations
 */
const listCampaigns = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let accountUUID = arguments.length > 1 ? arguments[1] : undefined;
  let trace = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  try {
    const MS = util.getEndpoint("campaigns");
    const requestOptions = {
      method: "GET",
      uri: `${MS}/10dlc/customers/${accountUUID}/campaigns`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "x-api-version": `${util.getVersion()}`,
        "Content-type": "application/json"
      },
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description - This function will remove numbers a SMS 10DLC Campaign
 * @param {string} [accessToken="null accessToken"] cpaas access token
 * @param {string} [accountUUID="null accountUUID"] cpaas account uuid to provision
 * @param {string} [campaignId="null campaignId"] 10dlc campaign id
 * @param {array} [numbers=[]] array of numbers to assign
 * @param {object} [trace={}] - microservice lifecyce headers
 * @returns {Promise} - Promise resolving to a object confirming the removal of numbers
 */
const removeFromSMSCampaign = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let accountUUID = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "null accountUUID";
  let campaignId = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : "null campaignId";
  let numbers = arguments.length > 3 ? arguments[3] : undefined;
  let trace = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : {};
  try {
    const MS = util.getEndpoint("campaigns");
    const requestOptions = {
      method: "POST",
      uri: `${MS}/10dlc/customers/${accountUUID}/campaigns/${campaignId}/numbers/unassign`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "x-api-version": `${util.getVersion()}`,
        "Content-type": "application/json"
      },
      body: {
        tns: numbers
      },
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function will update (re-register) a campaign for a CPaaS account that has an active brand registration
 * @param {string} [accessToken="null accessToken"] CPaaS access token
 * @param {string} [accountUUID="null accountUUID"] CPaaS account uuid
 * @param {string} [campaignId="null campaignId"] 10DLC campaign id to update
 * @param {string} description summary description of this campaign between 40 and 4096 characters
 * @param {string} displayName a friendly name for the campaign
 * @param {string} sampleMsg1 sample message 1
 * @param {string} sampleMsg2 sample message 2
 * @param {string} sampleMsg3 sample message 3
 * @param {string} sampleMsg4 sample message 4
 * @param {string} sampleMsg5 sample message 5
 * @param {boolean} numPooling campaign utilize pool of phone numbers?
 * @param {boolean} directLending messages used for lending or loan arangements?
 * @param {boolean} embeddedLink messages contain embedded URLs?
 * @param {boolean} embeddedPhone messages contain embedded phone numbers?
 * @param {boolean} affMarketing message content controlled by affiliate marketing other than the brand?
 * @param {boolean} ageGated campaign includes age gated message content?
 * @param {string} usecaseId usecase id
 * @param {array} subUsecaseId sub usecase ids (strings)
 * @param {string} [messageFlow="Users may opt-in by sending START to any number associated with the campaign. Users may also sign up to receive messages from this campaign via a website after accepting terms and conditions."] how subscribers are added or opt-in to the campaign
 * @param {string} termsLink url for viewing terms and conditions
 * @param {string} privacyLink url for viewing privacy policy
 * @param {boolean} [autoRenewal=true] subscription auto-renews?
 * @param {boolean} [subOptIn=true] provides automated opt-in?
 * @param {string} [optInKeywords="START"] opt-in keywords (comma separated)
 * @param {string} [optInMessage="You are now opted-in.\n\nFor help, reply HELP.\n\nTo opt-out, reply STOP"] automated opt-in message
 * @param {boolean} [subOptOut=true] provides automated opt-out?
 * @param {string} [optOutKeywords="STOP"] opt-out key words (comma separated)
 * @param {string} [optOutMessage="You have successfully opted-out.\n\nYou will not receive any more messages from this number.\n\nYou may reply START at any time to opt-in again."] automated opt-out message
 * @param {boolean} [subHelp=true] provides automated help?
 * @param {string} [helpKeywords="HELP"] help keywords (comma separated)
 * @param {string} [helpMessage="To opt-in and receive messages from this number, reply START.\n\nFor help, reply HELP.\n\nTo opt-out at any time, reply STOP"] automated help message
 * @param {object} [trace={}] optional CPaaS lifecycle headers
 * @return {Promsise<object>} promise resolving to a brand registration document
 */
const updateCampaign = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let accountUUID = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "null accountUUID";
  let campaignId = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : "null campaignId";
  let description = arguments.length > 3 ? arguments[3] : undefined;
  let displayName = arguments.length > 4 ? arguments[4] : undefined;
  let sampleMsg1 = arguments.length > 5 ? arguments[5] : undefined;
  let sampleMsg2 = arguments.length > 6 ? arguments[6] : undefined;
  let sampleMsg3 = arguments.length > 7 ? arguments[7] : undefined;
  let sampleMsg4 = arguments.length > 8 ? arguments[8] : undefined;
  let sampleMsg5 = arguments.length > 9 ? arguments[9] : undefined;
  let numPooling = arguments.length > 10 ? arguments[10] : undefined;
  let directLending = arguments.length > 11 ? arguments[11] : undefined;
  let embeddedLink = arguments.length > 12 ? arguments[12] : undefined;
  let embeddedPhone = arguments.length > 13 ? arguments[13] : undefined;
  let affMarketing = arguments.length > 14 ? arguments[14] : undefined;
  let ageGated = arguments.length > 15 ? arguments[15] : undefined;
  let usecaseId = arguments.length > 16 ? arguments[16] : undefined;
  let subUsecaseId = arguments.length > 17 ? arguments[17] : undefined;
  let messageFlow = arguments.length > 18 && arguments[18] !== undefined ? arguments[18] : "Users may opt-in by sending START to any number associated with the campaign. Users may also sign up to receive messages from this campaign via a website after accepting terms and conditions.";
  let termsLink = arguments.length > 19 ? arguments[19] : undefined;
  let privacyLink = arguments.length > 20 ? arguments[20] : undefined;
  let autoRenewal = arguments.length > 21 && arguments[21] !== undefined ? arguments[21] : true;
  let subOptIn = arguments.length > 22 && arguments[22] !== undefined ? arguments[22] : true;
  let optInKeywords = arguments.length > 23 && arguments[23] !== undefined ? arguments[23] : "START";
  let optInMessage = arguments.length > 24 && arguments[24] !== undefined ? arguments[24] : "You are now opted-in.\n\nFor help, reply HELP.\n\nTo opt-out, reply STOP";
  let subOptOut = arguments.length > 25 && arguments[25] !== undefined ? arguments[25] : true;
  let optOutKeywords = arguments.length > 26 && arguments[26] !== undefined ? arguments[26] : "STOP";
  let optOutMessage = arguments.length > 27 && arguments[27] !== undefined ? arguments[27] : "You have successfully opted-out.\n\nYou will not receive any more messages from this number.\n\nYou may reply START at any time to opt-in again.";
  let subHelp = arguments.length > 28 && arguments[28] !== undefined ? arguments[28] : true;
  let helpKeywords = arguments.length > 29 && arguments[29] !== undefined ? arguments[29] : "HELP";
  let helpMessage = arguments.length > 30 && arguments[30] !== undefined ? arguments[30] : "To opt-in and receive messages from this number, reply START.\n\nFor help, reply HELP.\n\nTo opt-out at any time, reply STOP";
  let trace = arguments.length > 31 && arguments[31] !== undefined ? arguments[31] : {};
  try {
    const MS = util.getEndpoint("campaigns");
    const requestOptions = {
      method: "PUT",
      uri: `${MS}/10dlc/customers/${accountUUID}/campaigns/${campaignId}`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "x-api-version": `${util.getVersion()}`,
        "Content-type": "application/json"
      },
      body: {
        description: description,
        display_name: displayName,
        message_flow: messageFlow,
        sample_msg1: sampleMsg1,
        sample_msg2: sampleMsg2,
        sample_msg3: sampleMsg3,
        sample_msg4: sampleMsg4,
        sample_msg5: sampleMsg5,
        terms_link: termsLink,
        privacy_link: privacyLink,
        sub_opt_in: subOptIn,
        sub_opt_out: subOptOut,
        sub_help: subHelp,
        num_pooling: numPooling,
        direct_lending: directLending,
        embedded_link: embeddedLink,
        embedded_phone: embeddedPhone,
        aff_marketing: affMarketing,
        age_gated: ageGated,
        usecase_id: usecaseId,
        sub_usecase_ids: subUsecaseId,
        help_message: helpMessage,
        auto_renewal: autoRenewal,
        opt_in_keywords: optInKeywords,
        opt_out_keywords: optOutKeywords,
        help_keywords: helpKeywords,
        opt_in_message: optInMessage,
        opt_out_message: optOutMessage
      },
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};
module.exports = {
  assignToSMSCampaign,
  createBrand,
  createCampaign,
  deleteBrand,
  deleteCampaign,
  getBrand,
  getBrandUseCases,
  getCampaign,
  getEnumerations,
  listCampaigns,
  removeFromSMSCampaign,
  updateCampaign
};