/* global require module*/
"use strict";

const request = require("./requestPromise");
const util = require("./utilities");
const logger = require("./node-logger").getInstance();

/**
 * @async
 * @description This function will add a subscription.
 * @param {string} [user_uuid="no user uuid provided"] - uuid for a star2star user
 * @param {string} [account_uuid="account uuid not provided "] - account to subscribe to
 * @param {string} [callback_url="not set callback"] - callback URL
 * @param {array}  [callback_headers=[]] - callback headers
 * @param {array}  criteria - optional filter criteria
 * @param {object} [subscriptions={}] - events to subscribe to (voice, fax, conferencing, messagin, sms,  presence)
 * @param {string} [accessToken="null accessToken"] - access token for cpaas systems
 * @param {string} [expiresDate=undefined] - optional expires date (RFC3339 format)
 * @param {object} [trace = {}] - optional microservice lifecycle trace headers
 * @param {boolean} [keep_alive = false] - optional high availability keep alive
 * @returns {Promise<object>} - Promise resolving to a data object containing a new subscription
 */
const addSubscription = async function () {
  let user_uuid = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "no user uuid provided";
  let account_uuid = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "account uuid not provided ";
  let callback_url = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : "not set callback";
  let callback_headers = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : [];
  let criteria = arguments.length > 4 ? arguments[4] : undefined;
  let subscriptions = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : {};
  let accessToken = arguments.length > 6 && arguments[6] !== undefined ? arguments[6] : "null accessToken";
  let expiresDate = arguments.length > 7 && arguments[7] !== undefined ? arguments[7] : undefined;
  let trace = arguments.length > 8 && arguments[8] !== undefined ? arguments[8] : {};
  let keep_alive = arguments.length > 9 && arguments[9] !== undefined ? arguments[9] : false;
  try {
    const MS = util.getEndpoint("pubsub");
    const requestOptions = {
      method: "POST",
      uri: `${MS}/subscriptions`,
      body: {
        user_uuid: user_uuid,
        account_uuid: account_uuid,
        callback: {
          url: callback_url,
          headers: callback_headers,
          keep_alive: keep_alive
        },
        //criteria: criteria, temporary sms workaround
        events: subscriptions
      },
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      json: true
    };

    // begin temporary sms workaround
    if (Array.isArray(criteria)) {
      let qs;
      const filteredCriteria = criteria.filter(elem => {
        return !(typeof elem === "object" && elem.hasOwnProperty("qs") && (qs = elem.qs));
      });
      if (typeof qs !== "undefined") {
        requestOptions.qs = qs;
      }
      if (filteredCriteria.length > 0) {
        requestOptions.body.criteria = filteredCriteria;
      }
    }
    // end temporary sms workaround

    if (expiresDate) {
      requestOptions.body.expiration_date = expiresDate;
    }
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function will add a subscription for a custom application and event.
 * @param {string} [accessToken="null accessToken"] - access token for cpaas systems
 * @param {string} [app_uuid="app_uuid uuid not provided "] - account to subscribe to
 * @param {string} [callback_url="not set callback"] - callback URL
 * @param {array}  [callback_headers=[]] - callback headers
 * @param {array}  criteria - optional filter criteria
 * @param {object} [events={}] - events to subscribe to (voice, fax, conferencing, messagin, sms,  presence)
 * @param {string} [expiresDate=undefined] - optional expires date (RFC3339 format)
 * @param {object} [trace = {}] - optional microservice lifecycle trace headers
 * @returns {Promise<object>} - Promise resolving to a data object containing a new subscription
 */
const addCustomEventSubscription = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let user_uuid = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "user uuid not provided";
  let app_uuid = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : "account uuid not provided ";
  let callback_url = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : "not set callback";
  let callback_headers = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : [];
  let criteria = arguments.length > 5 ? arguments[5] : undefined;
  let events = arguments.length > 6 && arguments[6] !== undefined ? arguments[6] : [];
  let expiresDate = arguments.length > 7 && arguments[7] !== undefined ? arguments[7] : undefined;
  let trace = arguments.length > 8 && arguments[8] !== undefined ? arguments[8] : {};
  try {
    const MS = util.getEndpoint("pubsub");
    const hasUserUuid = callback_headers.reduce((p, c) => {
      if (!p) {
        return c.hasOwnProperty("x-user-uuid");
      }
      return p;
    }, false);
    const user_headers = hasUserUuid === false ? [{
      "x-user-uuid": user_uuid
    }] : [];
    const requestOptions = {
      method: "POST",
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      uri: `${MS}/customevents`,
      body: {
        app_uuid: app_uuid,
        callback: {
          url: callback_url,
          headers: [...user_headers, ...callback_headers] // work around for CCORE-1545
        },
        events: events
      },
      json: true
    };
    if (Array.isArray(criteria) && criteria.length > 0) {
      requestOptions.body.criteria = criteria;
    }
    if (expiresDate) {
      requestOptions.body.expiration_date = expiresDate;
    }
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description - This function will send a message to subscribers of a custom application
 * @param {string} [accessToken="null accessToken"] - CPaaS access token
 * @param {string} [app_uuid="account uuid not provided "] - application uuid
 * @param {string} [event="null event"] - even string to broadcast
 * @param {object} [payload={}] - data to include in the broadcast
 * @param {object} [trace={}] - optional CPaaS lifecycle headers
 * @returns {Promise} - promise resolving to success or failure
 */
const broadcastCustomApplication = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let app_uuid = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "account uuid not provided ";
  let event = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : "null event";
  let payload = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};
  let trace = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : {};
  try {
    const MS = util.getEndpoint("pubsub");
    const requestOptions = {
      "method": "POST",
      "headers": {
        "Authorization": `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      "uri": `${MS}/broadcasts/applications`,
      "body": {
        "app_uuid": app_uuid,
        "event": event
      },
      "json": true
    };
    Object.keys(payload).forEach(prop => {
      requestOptions.body[prop] = payload[prop];
    });
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description - This function will create a custom pubsub application
 * @param {string} [accessToken="null accessToken"] - CPaaS access token
 * @param {string} [app_uuid="account uuid not provided "] - application_uuid
 * @param {array} [events=[]] - events as array of objects
 * @param {string} app_name - application name
 * @param {string} description - application description
 * @param {object} metadata - application metadata
 * @param {object} [trace={}] - optional CPaaS lifecycle headers
 * @returns {Promise} - promise resolving to success or failure
 */
const createCustomApplication = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let app_uuid = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "account uuid not provided ";
  let events = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : [];
  let app_name = arguments.length > 3 ? arguments[3] : undefined;
  let description = arguments.length > 4 ? arguments[4] : undefined;
  let metadata = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : {};
  let trace = arguments.length > 6 && arguments[6] !== undefined ? arguments[6] : {};
  try {
    const MS = util.getEndpoint("pubsub");
    const requestOptions = {
      "method": "POST",
      "headers": {
        "Authorization": `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      "uri": `${MS}/applications`,
      "body": {
        "app_uuid": app_uuid,
        "app_name": typeof app_name === "string" ? app_name : "",
        "description": typeof description === "string" ? description : "",
        "metadata": metadata,
        "events": events
      },
      "json": true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @desc - This function will delete a custom pubsub application
 * @param {string} [accessToken="null accessToken"] - CPaaS access token
 * @param {string} [app_uuid="account uuid not provided "] - application uuid
 * @param {*} [trace={}] - optional CPaaS lifecycle headers
 * @returns {Promise} - promise resolving to success or failure
 */
const deleteCustomApplication = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let app_uuid = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "account uuid not provided ";
  let trace = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  try {
    const MS = util.getEndpoint("pubsub");
    const requestOptions = {
      "method": "DELETE",
      "headers": {
        "Authorization": `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      "uri": `${MS}/applications/${app_uuid}`,
      "resolveWithFullResponse": true,
      "json": true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    if (response.hasOwnProperty("statusCode") && response.statusCode === 204) {
      return {
        "status": "ok"
      };
    } else {
      // this is an edge case, but protects against unexpected 2xx or 3xx response codes.
      throw {
        "code": response.statusCode,
        "message": typeof response.body === "string" ? response.body : "delete pubsub failed",
        "trace_id": requestOptions.hasOwnProperty("headers") && requestOptions.headers.hasOwnProperty("trace") ? requestOptions.headers.trace : undefined,
        "details": typeof response.body === "object" && response.body !== null ? [response.body] : []
      };
    }
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description - This function will delete a custom subscription
 * @param {string} [accessToken="null accessToken"]
 * @param {string} [subscription_uuid="no subscription uuid provided"]
 * @param {*} [trace={}]
 * @returns {Promise} - promise resolving to success or failure
 */
const deleteCustomSubscription = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let subscription_uuid = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "no subscription uuid provided";
  let trace = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  try {
    const MS = util.getEndpoint("pubsub");
    const requestOptions = {
      "method": "DELETE",
      "uri": `${MS}/customevents/${subscription_uuid}`,
      "headers": {
        "Authorization": `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      "resolveWithFullResponse": true,
      "json": true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    if (response.hasOwnProperty("statusCode") && response.statusCode === 204) {
      return {
        "status": "ok"
      };
    } else {
      // this is an edge case, but protects against unexpected 2xx or 3xx response codes.
      throw {
        "code": response.statusCode,
        "message": typeof response.body === "string" ? response.body : "delete pubsub failed",
        "trace_id": requestOptions.hasOwnProperty("headers") && requestOptions.headers.hasOwnProperty("trace") ? requestOptions.headers.trace : undefined,
        "details": typeof response.body === "object" && response.body !== null ? [response.body] : []
      };
    }
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function will delete a user subscriptions based on subscription id.
 * @param {string} [subscription_uuid="no subscription uuid provided"] - uuid for a star2star user
 * @param {string} [accessToken="null accessToken"] - access token for cpaas systems
 * @param {object} [trace = {}] - optional microservice lifecycle trace headers
 * @returns {Promise<empty>} - Promise resolving success or failure.
 */
const deleteSubscription = async function () {
  let subscription_uuid = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "no subscription uuid provided";
  let accessToken = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "null accessToken";
  let trace = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  try {
    const MS = util.getEndpoint("pubsub");
    const requestOptions = {
      method: "DELETE",
      uri: `${MS}/subscriptions/${subscription_uuid}`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      resolveWithFullResponse: true,
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    if (response.hasOwnProperty("statusCode") && response.statusCode === 204) {
      return {
        "status": "ok"
      };
    } else {
      // this is an edge case, but protects against unexpected 2xx or 3xx response codes.
      throw {
        "code": response.statusCode,
        "message": typeof response.body === "string" ? response.body : "delete pubsub failed",
        "trace_id": requestOptions.hasOwnProperty("headers") && requestOptions.headers.hasOwnProperty("trace") ? requestOptions.headers.trace : undefined,
        "details": typeof response.body === "object" && response.body !== null ? [response.body] : []
      };
    }
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description - This function will return a custom pubsub application
 * @param {string} [accessToken="null accessToken"] - CPaaS access token
 * @param {string} [app_uuid="account uuid not provided "] - application uuid
 * @param {*} [trace={}] - optional CPaaS lifecycle headers
 * @returns {Promise} - promise resolving to custom application object
 */
const getCustomApplication = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let app_uuid = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "account uuid not provided ";
  let trace = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  try {
    const MS = util.getEndpoint("pubsub");
    const requestOptions = {
      "method": "GET",
      "headers": {
        "Authorization": `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      "uri": `${MS}/applications/${app_uuid}`,
      "json": true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description - This function will return a custom subscription
 * @param {string} [accessToken="null accessToken"]
 * @param {string} [subscription_uuid="no subscription uuid provided"]
 * @param {*} [trace={}]
 * @returns {Promise} - promise resolving to subscription object
 */
const getCustomSubscription = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let subscription_uuid = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "no subscription uuid provided";
  let trace = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  try {
    const MS = util.getEndpoint("pubsub");
    const requestOptions = {
      method: "GET",
      uri: `${MS}/customevents/${subscription_uuid}`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function will get a user subscription based on subscription id.
 * @param {string} [subscription_uuid="no subscription uuid provided"] - uuid for a star2star user
 * @param {string} [accessToken="null accessToken"] - access token for cpaas systems
 * @param {object} [trace = {}] - optional microservice lifecycle trace headers
 * @returns {Promise<empty>} - Promise resolving success or failure.
 */
const getSubscription = async function () {
  let subscription_uuid = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "no subscription uuid provided";
  let accessToken = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "null accessToken";
  let trace = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  try {
    const MS = util.getEndpoint("pubsub");
    const requestOptions = {
      method: "GET",
      uri: `${MS}/subscriptions/${subscription_uuid}`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description - This function will return a custom subscription
 * @param {string} [accessToken="null accessToken"] - CPaaS access token
 * @param {string} [userUUID="no user uuid provided"] - custom user uuid
 * @param {string} [appUUID="no app uuid provided"] - custom application uuid
 * @param {number} [offset=0] - pagination offset
 * @param {number} [limit=10] - pagination limit
 * @param {object} [trace={}] - optional CPaaS lifecycle headers
 * @returns {Promise} - promise resolving to object containing items array of subscriptions
 */
const listCustomSubscriptions = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let userUUID = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "no user uuid provided";
  let appUUID = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : undefined;
  let offset = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : 0;
  let limit = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : 10;
  let trace = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : {};
  try {
    const MS = util.getEndpoint("pubsub");
    const requestOptions = {
      method: "GET",
      uri: `${MS}/customevents`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      qs: {},
      json: true
    };

    // add appUUID if set 
    if (appUUID) {
      requestOptions.qs.appUUID = appUUID;
    }
    util.addRequestTrace(requestOptions, trace);
    // work around for CCORE-1545
    const response = await util.aggregate(request, requestOptions, trace);
    const filteredResponse = util.filterResponse(response, {
      "x-user-uuid": userUUID
    });
    const paginatedResponse = util.paginate(filteredResponse, offset, limit);
    return paginatedResponse;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function will ask the cpaas pubsub service for the list of user's subscriptions.
 * @param {string} [accessToken="null accessToken"] - access token for cpaas systems
 * @param {string} [accountUUID="no account uuid provided"] - cpaas account uuid
 * @param {number} [offset=0] pagination offset
 * @param {number} [limit=10] pagination limit
 * @param {object} filters optional response filters in key-value
 * @param {object} [trace = {}] - optional microservice lifecycle trace headers
 * @returns {Promise<object>} - Promise resolving to a data object containing a list of subscriptions
 */
const listAccountSubscriptions = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let accountUUID = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "no user uuid provided";
  let offset = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 0;
  let limit = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : 10;
  let filters = arguments.length > 4 ? arguments[4] : undefined;
  let trace = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : {};
  try {
    const MS = util.getEndpoint("pubsub");
    const requestOptions = {
      method: "GET",
      uri: `${MS}/subscriptions`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      qs: {
        account_uuid: accountUUID,
        offset: offset,
        limit: limit
      },
      json: true
    };
    let response;
    if (typeof filters !== "undefined" && (typeof filters !== "object" || filters === null)) {
      // filter param has been passed in, make sure it is an array befor proceeding
      throw {
        "code": 400,
        "message": "filters param not an object",
        "trace_id": requestOptions.hasOwnProperty("headers") && requestOptions.headers.hasOwnProperty("trace") ? requestOptions.headers.trace : undefined,
        "details": [{
          "filters": filters
        }]
      };
    }
    if (typeof filters === "undefined" || Object.keys(filters).length === 0) {
      requestOptions.qs.offset = offset;
      requestOptions.qs.limit = limit;
      response = await request(requestOptions);
      return response;
    } else {
      response = await util.aggregate(request, requestOptions, trace);
      //logger.debug("****** AGGREGATE RESPONSE *******",response);
      if (response.hasOwnProperty("items") && response.items.length > 0) {
        const filteredResponse = util.filterResponse(response, filters);
        //logger.debug("******* FILTERED RESPONSE ********",filteredResponse);
        const paginatedResponse = util.paginate(filteredResponse, offset, limit);
        //logger.debug("******* PAGINATED RESPONSE ********",paginatedResponse);
        return paginatedResponse;
      } else {
        return response;
      }
    }
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function will ask the cpaas pubsub service for the list of user's subscriptions.
 * @param {string} [user_uuid="no user uuid provided"] - uuid for a star2star user
 * @param {string} [accessToken="null accessToken"] - access token for cpaas systems
 * @param {object} [trace = {}] - optional microservice lifecycle trace headers
 * @returns {Promise<object>} - Promise resolving to a data object containing a list of subscriptions for this user
 */
const listSubscriptions = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let offset = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0;
  let limit = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 10;
  let filters = arguments.length > 3 ? arguments[3] : undefined;
  let trace = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : {};
  try {
    const MS = util.getEndpoint("pubsub");
    const requestOptions = {
      method: "GET",
      uri: `${MS}/subscriptions`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      qs: {},
      json: true
    };
    let response;
    if (typeof filters !== "undefined" && (typeof filters !== "object" || filters === null)) {
      // filter param has been passed in, make sure it is an array befor proceeding
      throw {
        "code": 400,
        "message": "filters param not an object",
        "trace_id": requestOptions.hasOwnProperty("headers") && requestOptions.headers.hasOwnProperty("trace") ? requestOptions.headers.trace : undefined,
        "details": [{
          "filters": filters
        }]
      };
    }
    if (typeof filters === "undefined" || Object.keys(filters).length === 0) {
      requestOptions.qs.offset = offset;
      requestOptions.qs.limit = limit;
      response = await request(requestOptions);
      return response;
    } else {
      response = await util.aggregate(request, requestOptions, trace);
      // logger.debug("****** AGGREGATE RESPONSE *******",response);
      if (response.hasOwnProperty("items") && response.items.length > 0) {
        const filteredResponse = util.filterResponse(response, filters);
        // logger.debug("******* FILTERED RESPONSE ********",filteredResponse);
        const paginatedResponse = util.paginate(filteredResponse, offset, limit);
        //logger.debug("******* PAGINATED RESPONSE ********",paginatedResponse);
        return paginatedResponse;
      } else {
        return response;
      }
    }
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function will ask the cpaas pubsub service for the list of user's subscriptions.
 * @param {string} [user_uuid="no user uuid provided"] - uuid for a star2star user
 * @param {string} [accessToken="null accessToken"] - access token for cpaas systems
 * @param {object} [trace = {}] - optional microservice lifecycle trace headers
 * @returns {Promise<object>} - Promise resolving to a data object containing a list of subscriptions for this user
 */
const listUserSubscriptions = async function () {
  let user_uuid = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "no user uuid provided";
  let accessToken = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "null accessToken";
  let trace = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  try {
    const MS = util.getEndpoint("pubsub");
    const requestOptions = {
      method: "GET",
      uri: `${MS}/subscriptions`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      qs: {
        user_uuid: user_uuid
      },
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function updates a subscription
 * @param {string} [accessToken="null accessToken"] - access token for cpaas systems
 * @param {string} subscriptionUUID - subscription uuid
 * @param {object} [body="null body"] - subscription
 * @param {object} [trace={}] - optional cpaas lifecycle headers
 * @returns {Promise<object>} - Promise resolving to a data object containing updated subscription
 */
const updateSubscription = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let subscriptionUUID = arguments.length > 1 ? arguments[1] : undefined;
  let body = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : "null body";
  let trace = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};
  try {
    if (typeof body !== "object" || body === null) {
      throw {
        code: 400,
        message: "request body invalid format"
      };
    }
    const MS = util.getEndpoint("pubsub");
    const requestOptions = {
      method: "PUT",
      uri: `${MS}/subscriptions/${subscriptionUUID}`,
      body: body,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function updates a subscription expiration date
 * @param {string} [accessToken="null accessToken"] - access token for cpaas systems
 * @param {*} subscriptionUUID - uuid for a star2star user
 * @param {*} [expiresDate=new Date(Date.now()).toISOString()]
 * @param {*} [trace={}]
 * @returns {Promise<object>} - Promise resolving to a data object containing updated subscription
 */
const updateSubscriptionExpiresDate = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let subscriptionUUID = arguments.length > 1 ? arguments[1] : undefined;
  let expiresDate = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : new Date(Date.now()).toISOString();
  let trace = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};
  try {
    const MS = util.getEndpoint("pubsub");
    const requestOptions = {
      method: "PUT",
      uri: `${MS}/subscriptions/${subscriptionUUID}`,
      body: {
        expiration_date: expiresDate
      },
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      json: true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description - This function will return a list of all custom pubsub applications
 * @param {string} [accessToken="null accessToken"] - CPaaS access token
 * @param {*} [trace={}] - optional CPaaS lifecycle headers
 * @returns {Promise} - promise resolving to custom application object
 */
const listCustomApplications = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let app_uuid = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "account uuid not provided ";
  let trace = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  try {
    const MS = util.getEndpoint("pubsub");
    const requestOptions = {
      "method": "GET",
      "headers": {
        "Authorization": `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      "uri": `${MS}/applications`,
      "json": true
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};
module.exports = {
  addSubscription,
  addCustomEventSubscription,
  broadcastCustomApplication,
  createCustomApplication,
  deleteCustomApplication,
  deleteCustomSubscription,
  deleteSubscription,
  getCustomApplication,
  getCustomSubscription,
  getSubscription,
  listCustomApplications,
  listCustomSubscriptions,
  listAccountSubscriptions,
  listSubscriptions,
  listUserSubscriptions,
  updateSubscription,
  updateSubscriptionExpiresDate
};