/*global module require */
"use strict";

const util = require("./utilities");
const request = require("./requestPromise");

/**
 * @async
 * @description This function creates a new usage event.
 * @param {string} [accessToken="null accessToken"] - access token for cpaas systems
 * @param {string} [accountUUID="undefined"] - account UUID
 * @param {string} [applicationUUID="undefined"] - application UUID
 * @param {string} [eventType="undefined"] - event type
 * @param {string} [metadata="undefined"] - object containing metadata for this event
 * @param {object} [trace = {}] - optional microservice lifecycle trace headers
 * @returns {Promise<object>} - Promise resolving to the created usage event object
 */
const createEvent = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let accountUUID = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : undefined;
  let applicationUUID = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : undefined;
  let eventType = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : undefined;
  let metadata = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : undefined;
  let trace = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : {};
  try {
    const MS = util.getEndpoint("usage");
    const requestOptions = {
      method: "POST",
      uri: `${MS}/events`,
      body: {
        account_uuid: accountUUID,
        application_uuid: applicationUUID,
        event_type: eventType,
        metadata: metadata
      },
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      json: true,
      resolveWithFullResponse: false
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function get a CPaaS account billing cycle.
 * @param {string} [accessToken="null accessToken"] - access token for cpaas systems
 * @param {string} [accountUUID="undefined"] - account UUID
 * @param {object} [trace = {}] - optional microservice lifecycle trace headers
 * @returns {Promise<object>} - Promise resolving to the created usage event object
 */
const getAccountBillingCycle = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let accountUUID = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : undefined;
  let trace = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  try {
    const MS = util.getEndpoint("usage");
    const requestOptions = {
      method: "GET",
      uri: `${MS}/accounts/${accountUUID}/billing_cycle`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      json: true,
      resolveWithFullResponse: false
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function gets a completed usage report.
 * @param {string} [accessToken="null accessToken"] - access token for cpaas systems
 * @param {string} [reportUUID="undefined"] - repor UUID
 * @param {object} [trace = {}] - optional microservice lifecycle trace headers
 * @returns {Promise<object>} - Promise resolving to an object containing report location and details
 */
const getUsageReport = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let reportUUID = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : undefined;
  let trace = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  try {
    const MS = util.getEndpoint("usage");
    const requestOptions = {
      method: "GET",
      uri: `${MS}/reports/${reportUUID}`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      json: true,
      resolveWithFullResponse: false
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function gets a usage template by uuid.
 * @param {string} [accessToken="null accessToken"] - access token for cpaas systems
 * @param {string} [templateUUID="null templateUUID"] - usage template UUID
 * @param {object} [trace = {}] - optional microservice lifecycle trace headers
 * @returns {Promise<object>} - Promise resolving to a usage report template
 */
const getUsageTemplate = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let templateUUID = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "null templateUUID";
  let trace = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  try {
    const MS = util.getEndpoint("usage");
    const requestOptions = {
      method: "GET",
      uri: `${MS}/reports/templates/${templateUUID}`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      json: true,
      resolveWithFullResponse: false
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function lists available usage report templates.
 * @param {string} [accessToken="null accessToken"] - access token for cpaas systems
 * @param {string} [offset=0] - pagination offset
 * @param {string} [limit=0] - pagination limit
 * @param {object} [trace = {}] - optional microservice lifecycle trace headers
 * @returns {Promise<object>} - Promise resolving to a list of report templates
 */
const listUsageTemplates = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let offset = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0;
  let limit = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 100;
  let trace = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};
  try {
    const MS = util.getEndpoint("usage");
    const requestOptions = {
      method: "GET",
      uri: `${MS}/reports/templates`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      qs: {
        offset: offset,
        limit: limit
      },
      json: true,
      resolveWithFullResponse: false
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};

/**
 * @async
 * @description This function starts a usage report.
 * @param {string} [accessToken="null accessToken"] - access token for cpaas systems
 * @param {string} [templateUUID="null templateUUID"] - report template uuid UUID
 * @param {string} [since="null since"] - report start
 * @param {string} [until="null until"] - report end
 * @param {string} accountUUID - optional CPaaS account uuid
 * @param {string} applicationUUID - optional CPaaS application uuid
 * @param {string} eventType - optional event type
 * @param {number} offset - optional offset
 * @param {number} limit - optional limit
 * @param {object} filterValues - optional filters expression object
 * @param {object} [trace = {}] - optional microservice lifecycle trace headers
 * @returns {Promise<object>} - Promise resolving to the created usage report metadata
 */
const runUsageReport = async function () {
  let accessToken = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "null accessToken";
  let templateUUID = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "null templateUUID";
  let since = arguments.length > 2 ? arguments[2] : undefined;
  let until = arguments.length > 3 ? arguments[3] : undefined;
  let accountUUID = arguments.length > 4 ? arguments[4] : undefined;
  let applicationUUID = arguments.length > 5 ? arguments[5] : undefined;
  let eventType = arguments.length > 6 ? arguments[6] : undefined;
  let offset = arguments.length > 7 ? arguments[7] : undefined;
  let limit = arguments.length > 8 ? arguments[8] : undefined;
  let filterValues = arguments.length > 9 ? arguments[9] : undefined;
  let trace = arguments.length > 10 && arguments[10] !== undefined ? arguments[10] : {};
  try {
    const MS = util.getEndpoint("usage");
    const requestOptions = {
      method: "POST",
      uri: `${MS}/reports/templates/${templateUUID}/events/evaluate`,
      headers: {
        Authorization: `Bearer ${accessToken}`,
        "Content-type": "application/json",
        "x-api-version": `${util.getVersion()}`
      },
      body: {
        account_uuid: accountUUID,
        application_uuid: applicationUUID,
        event_type: eventType,
        since: since,
        until: until,
        offset: offset,
        limit: limit,
        filter_values: filterValues
      },
      json: true,
      resolveWithFullResponse: false
    };
    util.addRequestTrace(requestOptions, trace);
    const response = await request(requestOptions);
    return response;
  } catch (error) {
    throw util.formatError(error);
  }
};
module.exports = {
  createEvent,
  getAccountBillingCycle,
  getUsageReport,
  getUsageTemplate,
  listUsageTemplates,
  runUsageReport
};